@@ -13,12 +13,10 @@
 #include "erofs/print.h"
 #include "erofs/inode.h"
 #include "erofs/dir.h"
-#include "erofs/compress.h"
-#include "erofs/fragments.h"
+#include "../lib/liberofs_compress.h"
 #include "../lib/liberofs_private.h"
 #include "../lib/liberofs_uuid.h"
 
-
 struct erofsdump_cfg {
 	unsigned int totalshow;
 	bool show_inode;
@@ -27,6 +25,7 @@ struct erofsdump_cfg {
 	bool show_statistics;
 	bool show_subdirectories;
 	bool show_file_content;
+	bool show_blkid_udev;
 	erofs_nid_t nid;
 	const char *inode_path;
 };
@@ -82,28 +81,32 @@ static struct option long_options[] = {
 	{"ls", no_argument, NULL, 5},
 	{"offset", required_argument, NULL, 6},
 	{"cat", no_argument, NULL, 7},
+	{"blkid-udev", no_argument, NULL, 512},
 	{0, 0, 0, 0},
 };
 
 struct erofsdump_feature {
 	bool compat;
+	u16 lkver;
 	u32 flag;
 	const char *name;
 };
 
 static struct erofsdump_feature feature_lists[] = {
-	{ true, EROFS_FEATURE_COMPAT_SB_CHKSUM, "sb_csum" },
-	{ true, EROFS_FEATURE_COMPAT_MTIME, "mtime" },
-	{ true, EROFS_FEATURE_COMPAT_XATTR_FILTER, "xattr_filter" },
-	{ false, EROFS_FEATURE_INCOMPAT_ZERO_PADDING, "0padding" },
-	{ false, EROFS_FEATURE_INCOMPAT_COMPR_CFGS, "compr_cfgs" },
-	{ false, EROFS_FEATURE_INCOMPAT_BIG_PCLUSTER, "big_pcluster" },
-	{ false, EROFS_FEATURE_INCOMPAT_CHUNKED_FILE, "chunked_file" },
-	{ false, EROFS_FEATURE_INCOMPAT_DEVICE_TABLE, "device_table" },
-	{ false, EROFS_FEATURE_INCOMPAT_ZTAILPACKING, "ztailpacking" },
-	{ false, EROFS_FEATURE_INCOMPAT_FRAGMENTS, "fragments" },
-	{ false, EROFS_FEATURE_INCOMPAT_DEDUPE, "dedupe" },
-	{ false, EROFS_FEATURE_INCOMPAT_XATTR_PREFIXES, "xattr_prefixes" },
+	{  true,    0, EROFS_FEATURE_COMPAT_SB_CHKSUM, "sb_csum" },
+	{  true,    0, EROFS_FEATURE_COMPAT_MTIME, "mtime" },
+	{  true,    0, EROFS_FEATURE_COMPAT_XATTR_FILTER, "xattr_filter" },
+	{ false, 504U, EROFS_FEATURE_INCOMPAT_LZ4_0PADDING, "lz4_0padding" },
+	{ false, 513U, EROFS_FEATURE_INCOMPAT_COMPR_CFGS, "compr_cfgs" },
+	{ false, 513U, EROFS_FEATURE_INCOMPAT_BIG_PCLUSTER, "big_pcluster" },
+	{ false, 515U, EROFS_FEATURE_INCOMPAT_CHUNKED_FILE, "chunked_file" },
+	{ false, 516U, EROFS_FEATURE_INCOMPAT_DEVICE_TABLE, "device_table" },
+	{ false, 517U, EROFS_FEATURE_INCOMPAT_ZTAILPACKING, "ztailpacking" },
+	{ false, 601U, EROFS_FEATURE_INCOMPAT_FRAGMENTS, "fragments" },
+	{ false, 601U, EROFS_FEATURE_INCOMPAT_DEDUPE, "dedupe" },
+	{ false, 604U, EROFS_FEATURE_INCOMPAT_XATTR_PREFIXES, "xattr_prefixes" },
+	{ false, 615U, EROFS_FEATURE_INCOMPAT_48BIT, "48bit" },
+	{ false, 617U, EROFS_FEATURE_INCOMPAT_METABOX, "metabox" },
 };
 
 static int erofsdump_readdir(struct erofs_dir_context *ctx);
@@ -123,6 +126,7 @@ static void usage(int argc, char **argv)
 		" -S              show statistic information of the image\n"
 		" -e              show extent info (INODE required)\n"
 		" -s              show information about superblock\n"
+		" --blkid-udev    print block device attributes for easy import into the udev environment\n"
 		" --device=X      specify an extra device to be used together\n"
 		" --ls            show directory contents (INODE required)\n"
 		" --cat           show file contents (INODE required)\n"
@@ -162,7 +166,11 @@ static int erofsdump_parse_options_cfg(int argc, char **argv)
 			exit(0);
 		case 2:
 			dumpcfg.show_inode = true;
-			dumpcfg.nid = (erofs_nid_t)atoll(optarg);
+			dumpcfg.nid = strtoull(optarg, &endptr, 0);
+			if (*endptr != '\0') {
+				erofs_err("invalid NID %s", optarg);
+				return -EINVAL;
+			}
 			++dumpcfg.totalshow;
 			break;
 		case 'h':
@@ -192,6 +200,9 @@ static int erofsdump_parse_options_cfg(int argc, char **argv)
 		case 7:
 			dumpcfg.show_file_content = true;
 			break;
+		case 512:
+			dumpcfg.show_blkid_udev = true;
+			break;
 		default:
 			return -EINVAL;
 		}
@@ -373,8 +384,9 @@ static void erofsdump_show_fileinfo(bool show_extent)
 	char access_mode_str[] = "rwxrwxrwx";
 	char timebuf[128] = {0};
 	unsigned int extent_count = 0;
+	struct tm *tm;
 	struct erofs_map_blocks map = {
-		.index = UINT_MAX,
+		.buf = __EROFS_BUF_INITIALIZER,
 		.m_la = 0,
 	};
 
@@ -405,8 +417,11 @@ static void erofsdump_show_fileinfo(bool show_extent)
 		path[sizeof(path) - 1] = '\0';
 	}
 
-	strftime(timebuf, sizeof(timebuf),
-		 "%Y-%m-%d %H:%M:%S", localtime((time_t *)&inode.i_mtime));
+	tm = localtime((time_t *)&inode.i_mtime);
+	if (!tm)
+		sprintf(timebuf, "%lld", (s64)inode.i_mtime | 0LL);
+	else
+		strftime(timebuf, sizeof(timebuf), "%Y-%m-%d %H:%M:%S", tm);
 	access_mode = inode.i_mode & 0777;
 	for (i = 8; i >= 0; i--)
 		if (((access_mode >> i) & 1) == 0)
@@ -623,9 +638,10 @@ static void erofsdump_print_supported_compressors(FILE *f, unsigned int mask)
 
 static void erofsdump_show_superblock(void)
 {
-	time_t time = g_sbi.build_time;
+	time_t time = g_sbi.epoch + g_sbi.build_time;
+	unsigned int minkver = 0;
 	char uuid_str[37];
-	int i = 0;
+	int i;
 
 	fprintf(stdout, "Filesystem magic number:                      0x%04X\n",
 			EROFS_SUPER_MAGIC_V1);
@@ -642,6 +658,9 @@ static void erofsdump_show_superblock(void)
 	if (erofs_sb_has_fragments(&g_sbi) && g_sbi.packed_nid > 0)
 		fprintf(stdout, "Filesystem packed nid:                        %llu\n",
 			g_sbi.packed_nid | 0ULL);
+	if (erofs_sb_has_metabox(&g_sbi))
+		fprintf(stdout, "Filesystem metabox nid:                       %llu\n",
+			g_sbi.metabox_nid | 0ULL);
 	if (erofs_sb_has_compr_cfgs(&g_sbi)) {
 		fprintf(stdout, "Filesystem compr_algs:                        ");
 		erofsdump_print_supported_compressors(stdout,
@@ -657,16 +676,30 @@ static void erofsdump_show_superblock(void)
 	fprintf(stdout, "Filesystem created:                           %s",
 			ctime(&time));
 	fprintf(stdout, "Filesystem features:                          ");
-	for (; i < ARRAY_SIZE(feature_lists); i++) {
+	for (i = 0; i < ARRAY_SIZE(feature_lists); i++) {
 		u32 feat = le32_to_cpu(feature_lists[i].compat ?
 				       g_sbi.feature_compat :
 				       g_sbi.feature_incompat);
-		if (feat & feature_lists[i].flag)
+		if (feat & feature_lists[i].flag) {
 			fprintf(stdout, "%s ", feature_lists[i].name);
+			if (feature_lists[i].lkver > minkver)
+				minkver = feature_lists[i].lkver;
+		}
 	}
 	erofs_uuid_unparse_lower(g_sbi.uuid, uuid_str);
-	fprintf(stdout, "\nFilesystem UUID:                              %s\n",
-			uuid_str);
+	fprintf(stdout, "\nFilesystem UUID:                              %s",
+		uuid_str);
+	if ((g_sbi.available_compr_algs >> Z_EROFS_COMPRESSION_LZMA) & 1)
+		minkver = max(minkver, 516U);
+	if ((g_sbi.available_compr_algs >> Z_EROFS_COMPRESSION_DEFLATE) & 1)
+		minkver = max(minkver, 606U);
+	if ((g_sbi.available_compr_algs >> Z_EROFS_COMPRESSION_ZSTD) & 1)
+		minkver = max(minkver, 610U);
+	if (minkver)
+		fprintf(stdout, "\nRequired upstream Linux kernel version:       %u.%u\n",
+			minkver / 100, minkver % 100);
+	else
+		fputs("\n", stdout);
 }
 
 static void erofsdump_show_file_content(void)
@@ -678,6 +711,7 @@ static void erofsdump_show_file_content(void)
 	erofs_off_t pending_size;
 	erofs_off_t read_offset;
 	erofs_off_t read_size;
+	struct erofs_vfile vf;
 
 	if (dumpcfg.inode_path) {
 		err = erofs_ilookup(dumpcfg.inode_path, &inode);
@@ -693,6 +727,10 @@ static void erofsdump_show_file_content(void)
 		}
 	}
 
+	err = erofs_iopen(&vf, &inode);
+	if (err)
+		return;
+
 	buffer_size = erofs_blksiz(inode.sbi);
 	buffer_ptr = malloc(buffer_size);
 	if (!buffer_ptr) {
@@ -703,8 +741,9 @@ static void erofsdump_show_file_content(void)
 	pending_size = inode.i_size;
 	read_offset = 0;
 	while (pending_size > 0) {
-		read_size = pending_size > buffer_size? buffer_size: pending_size;
-		err = erofs_pread(&inode, buffer_ptr, read_size, read_offset);
+		read_size = pending_size > buffer_size ?
+			buffer_size : pending_size;
+		err = erofs_pread(&vf, buffer_ptr, read_size, read_offset);
 		if (err) {
 			erofs_err("read file failed @ nid %llu", inode.nid | 0ULL);
 			goto out;
@@ -731,6 +770,9 @@ int main(int argc, char **argv)
 		goto exit;
 	}
 
+	if (dumpcfg.show_blkid_udev)
+		cfg.c_dbg_lvl = -1;
+
 	err = erofs_dev_open(&g_sbi, cfg.c_img_path, O_RDONLY | O_TRUNC);
 	if (err) {
 		erofs_err("failed to open image file");
@@ -743,6 +785,15 @@ int main(int argc, char **argv)
 		goto exit_dev_close;
 	}
 
+	if (dumpcfg.show_blkid_udev) {
+		char uuid_str[37];
+
+		erofs_uuid_unparse_lower(g_sbi.uuid, uuid_str);
+		printf("ID_FS_UUID=%s\nID_FS_UUID_ENC=%s\nID_FS_TYPE=erofs\nID_FS_USAGE=filesystem\n",
+		       uuid_str, uuid_str);
+		goto exit_put_super;
+	}
+
 	if (dumpcfg.show_file_content) {
 		if (dumpcfg.show_superblock || dumpcfg.show_statistics || dumpcfg.show_subdirectories) {
 			fprintf(stderr, "The '--cat' flag is incompatible with '-S', '-e', '-s' and '--ls'.\n");
