@@ -88,6 +88,10 @@ typedef int64_t         s64;
 #define le32_to_cpu(x) ((__u32)(x))
 #define le64_to_cpu(x) ((__u64)(x))
 
+#define cpu_to_be32(x) ((__be32)__builtin_bswap32(x))
+#define cpu_to_be64(x) ((__be64)__builtin_bswap64(x))
+#define be32_to_cpu(x) (__builtin_bswap32(x))
+#define be64_to_cpu(x) (__builtin_bswap64(x))
 #else
 #if __BYTE_ORDER == __BIG_ENDIAN
 #define cpu_to_le16(x) (__builtin_bswap16(x))
@@ -96,6 +100,11 @@ typedef int64_t         s64;
 #define le16_to_cpu(x) (__builtin_bswap16(x))
 #define le32_to_cpu(x) (__builtin_bswap32(x))
 #define le64_to_cpu(x) (__builtin_bswap64(x))
+
+#define cpu_to_be32(x) ((__be32)(x))
+#define cpu_to_be64(x) ((__be64)(x))
+#define be32_to_cpu(x) ((__u32)(x))
+#define be64_to_cpu(x) ((__u64)(x))
 #else
 #pragma error
 #endif
@@ -369,6 +378,15 @@ unsigned long __roundup_pow_of_two(unsigned long n)
 #define __erofs_likely(x)      __builtin_expect(!!(x), 1)
 #define __erofs_unlikely(x)    __builtin_expect(!!(x), 0)
 
+#if __has_attribute(__fallthrough__)
+# define __erofs_fallthrough	__attribute__((__fallthrough__))
+#else
+# define __erofs_fallthrough	do {} while (0)  /* fallthrough */
+#endif
+
+#define __erofs_stringify_1(x...)	#x
+#define __erofs_stringify(x...)		__erofs_stringify_1(x)
+
 #ifdef __cplusplus
 }
 #endif
