@@ -16,6 +16,9 @@ static inline void erofs_mutex_init(erofs_mutex_t *lock)
 #define erofs_mutex_lock	pthread_mutex_lock
 #define erofs_mutex_unlock	pthread_mutex_unlock
 
+#define EROFS_DEFINE_MUTEX(lock)	\
+	erofs_mutex_t lock = PTHREAD_MUTEX_INITIALIZER
+
 typedef pthread_rwlock_t erofs_rwsem_t;
 
 static inline void erofs_init_rwsem(erofs_rwsem_t *lock)
@@ -33,6 +36,9 @@ static inline void erofs_mutex_init(erofs_mutex_t *lock) {}
 static inline void erofs_mutex_lock(erofs_mutex_t *lock) {}
 static inline void erofs_mutex_unlock(erofs_mutex_t *lock) {}
 
+#define EROFS_DEFINE_MUTEX(lock)	\
+	erofs_mutex_t lock = {}
+
 typedef struct {} erofs_rwsem_t;
 static inline void erofs_init_rwsem(erofs_rwsem_t *lock) {}
 static inline void erofs_down_read(erofs_rwsem_t *lock) {}
