@@ -85,8 +85,27 @@ int erofs_compress_destsize(const struct erofs_compress *c,
 	return c->alg->c->compress_destsize(c, src, srcsize, dst, dstsize);
 }
 
+int erofs_compress(const struct erofs_compress *c,
+		   const void *src, unsigned int srcsize,
+		   void *dst, unsigned int dstcapacity)
+{
+	DBG_BUGON(!c->alg);
+	if (!c->alg->c->compress)
+		return -EOPNOTSUPP;
+
+	return c->alg->c->compress(c, src, srcsize, dst, dstcapacity);
+}
+
+int erofs_compressor_exit(struct erofs_compress *c)
+{
+	if (c->alg && c->alg->c->exit)
+		return c->alg->c->exit(c);
+	return 0;
+}
+
 int erofs_compressor_init(struct erofs_sb_info *sbi, struct erofs_compress *c,
-			  char *alg_name, int compression_level, u32 dict_size)
+			  const struct z_erofs_paramset *zset,
+			  u32 pclustersize_max)
 {
 	int ret, i;
 
@@ -97,63 +116,76 @@ int erofs_compressor_init(struct erofs_sb_info *sbi, struct erofs_compress *c,
 	c->compression_level = -1;
 	c->dict_size = 0;
 
-	if (!alg_name) {
-		c->alg = NULL;
-		return 0;
-	}
-
 	ret = -EINVAL;
 	for (i = 0; i < ARRAY_SIZE(erofs_algs); ++i) {
-		if (alg_name && strcmp(alg_name, erofs_algs[i].name))
+		if (strcmp(zset->alg, erofs_algs[i].name))
 			continue;
 
 		if (!erofs_algs[i].c)
 			continue;
 
+		if (!erofs_algs[i].c->setlevel && zset->clevel >= 0) {
+			erofs_err("compression level %d is not supported for %s",
+				  zset->clevel, zset->alg);
+			return -EINVAL;
+		}
+
+		if (!erofs_algs[i].c->setdictsize && zset->dict_size) {
+			erofs_err("unsupported dict size for %s", zset->alg);
+			return -EINVAL;
+		}
+
+		if (!erofs_algs[i].c->setextraopts && zset->extraopts) {
+			erofs_err("invalid compression option %s for %s",
+				  zset->extraopts, zset->alg);
+			return -EINVAL;
+		}
+
+		if (erofs_algs[i].c->preinit) {
+			ret = erofs_algs[i].c->preinit(c);
+			if (ret)
+				return ret;
+		}
+
 		if (erofs_algs[i].c->setlevel) {
-			ret = erofs_algs[i].c->setlevel(c, compression_level);
+			ret = erofs_algs[i].c->setlevel(c, zset->clevel);
 			if (ret) {
 				erofs_err("failed to set compression level %d for %s",
-					  compression_level, alg_name);
-				return ret;
+					  zset->clevel, zset->alg);
+				goto fail;
 			}
-		} else if (compression_level >= 0) {
-			erofs_err("compression level %d is not supported for %s",
-				  compression_level, alg_name);
-			return -EINVAL;
 		}
 
 		if (erofs_algs[i].c->setdictsize) {
-			ret = erofs_algs[i].c->setdictsize(c, dict_size);
+			ret = erofs_algs[i].c->setdictsize(c, zset->dict_size,
+							   pclustersize_max);
 			if (ret) {
 				erofs_err("failed to set dict size %u for %s",
-					  dict_size, alg_name);
-				return ret;
+					  zset->dict_size, zset->alg);
+				goto fail;
 			}
-		} else if (dict_size) {
-			erofs_err("dict size is not supported for %s",
-				  alg_name);
-			return -EINVAL;
 		}
 
-		ret = erofs_algs[i].c->init(c);
-		if (ret)
-			return ret;
-
-		if (!ret) {
-			c->alg = &erofs_algs[i];
-			return 0;
+		if (zset->extraopts && erofs_algs[i].c->setextraopts) {
+			ret = erofs_algs[i].c->setextraopts(c, zset->extraopts);
+			if (ret)
+				goto fail;
 		}
+
+		if (erofs_algs[i].c->init) {
+			ret = erofs_algs[i].c->init(c);
+			if (ret)
+				goto fail;
+		}
+		c->alg = &erofs_algs[i];
+		return 0;
 	}
-	erofs_err("Cannot find a valid compressor %s", alg_name);
+	erofs_err("Cannot find a valid compressor %s", zset->alg);
+	return ret;
+fail:
+	if (erofs_algs[i].c->preinit && erofs_algs[i].c->exit)
+		erofs_algs[i].c->exit(c);
 	return ret;
-}
-
-int erofs_compressor_exit(struct erofs_compress *c)
-{
-	if (c->alg && c->alg->c->exit)
-		return c->alg->c->exit(c);
-	return 0;
 }
 
 void erofs_compressor_reset(struct erofs_compress *c)
