@@ -37,8 +37,6 @@ static int compressor_deflate_exit(struct erofs_compress *c)
 
 static int compressor_deflate_init(struct erofs_compress *c)
 {
-	static erofs_atomic_bool_t __warnonce;
-
 	if (c->private_data) {
 		kite_deflate_end(c->private_data);
 		c->private_data = NULL;
@@ -46,12 +44,6 @@ static int compressor_deflate_init(struct erofs_compress *c)
 	c->private_data = kite_deflate_init(c->compression_level, c->dict_size);
 	if (IS_ERR_VALUE(c->private_data))
 		return PTR_ERR(c->private_data);
-
-	if (!erofs_atomic_test_and_set(&__warnonce)) {
-		erofs_warn("EXPERIMENTAL DEFLATE algorithm in use. Use at your own risk!");
-		erofs_warn("*Carefully* check filesystem data correctness to avoid corruption!");
-		erofs_warn("Please send a report to <linux-erofs@lists.ozlabs.org> if something is wrong.");
-	}
 	return 0;
 }
 
@@ -70,7 +62,8 @@ static int erofs_compressor_deflate_setlevel(struct erofs_compress *c,
 }
 
 static int erofs_compressor_deflate_setdictsize(struct erofs_compress *c,
-						u32 dict_size)
+						u32 dict_size,
+						u32 pclustersize_max)
 {
 	if (!dict_size)
 		dict_size = erofs_compressor_deflate.default_dictsize;
