@@ -50,32 +50,58 @@ static int erofs_compressor_liblzma_exit(struct erofs_compress *c)
 
 	lzma_end(&ctx->strm);
 	free(ctx);
+	c->private_data = NULL;
+	return 0;
+}
+
+static int erofs_compressor_liblzma_preinit(struct erofs_compress *c)
+{
+	struct erofs_liblzma_context *ctx;
+
+	ctx = malloc(sizeof(*ctx));
+	if (!ctx)
+		return -ENOMEM;
+	ctx->strm = (lzma_stream)LZMA_STREAM_INIT;
+	DBG_BUGON(c->private_data);
+	c->private_data = ctx;
 	return 0;
 }
 
 static int erofs_compressor_liblzma_setlevel(struct erofs_compress *c,
 					     int compression_level)
 {
-	if (compression_level < 0)
-		compression_level = erofs_compressor_lzma.default_level;
+	struct erofs_liblzma_context *ctx = c->private_data;
+	u32 preset;
 
 	if (compression_level > erofs_compressor_lzma.best_level) {
 		erofs_err("invalid compression level %d", compression_level);
 		return -EINVAL;
 	}
+
+	if (compression_level < 0)
+		preset = LZMA_PRESET_DEFAULT;
+	else if (compression_level >= 100)
+		preset = (compression_level - 100) | LZMA_PRESET_EXTREME;
+	else
+		preset = compression_level;
+
+	if (lzma_lzma_preset(&ctx->opt, preset))
+		return -EINVAL;
 	c->compression_level = compression_level;
 	return 0;
 }
 
 static int erofs_compressor_liblzma_setdictsize(struct erofs_compress *c,
-						u32 dict_size)
+						u32 dict_size, u32 pclustersize_max)
 {
+	struct erofs_liblzma_context *ctx = c->private_data;
+
 	if (!dict_size) {
 		if (erofs_compressor_lzma.default_dictsize) {
 			dict_size = erofs_compressor_lzma.default_dictsize;
 		} else {
 			dict_size = min_t(u32, Z_EROFS_LZMA_MAX_DICT_SIZE,
-					  cfg.c_mkfs_pclustersize_max << 2);
+					  pclustersize_max << 2);
 			if (dict_size < 32768)
 				dict_size = 32768;
 		}
@@ -85,32 +111,52 @@ static int erofs_compressor_liblzma_setdictsize(struct erofs_compress *c,
 		erofs_err("invalid dictionary size %u", dict_size);
 		return -EINVAL;
 	}
-	c->dict_size = dict_size;
+	ctx->opt.dict_size = c->dict_size = dict_size;
 	return 0;
 }
 
-static int erofs_compressor_liblzma_init(struct erofs_compress *c)
+static int erofs_compressor_liblzma_setextraopts(struct erofs_compress *c,
+						 const char *extraopts)
 {
-	struct erofs_liblzma_context *ctx;
-	u32 preset;
+	struct erofs_liblzma_context *ctx = c->private_data;
+	const char *token, *next;
 
-	ctx = malloc(sizeof(*ctx));
-	if (!ctx)
-		return -ENOMEM;
-	ctx->strm = (lzma_stream)LZMA_STREAM_INIT;
+	for (token = extraopts; *token != '\0'; token = next) {
+		const char *p = strchr(token, ',');
+		const char *rhs;
+		char *endptr;
+		unsigned long val;
+		uint32_t *key;
 
-	if (c->compression_level < 0)
-		preset = LZMA_PRESET_DEFAULT;
-	else if (c->compression_level >= 100)
-		preset = (c->compression_level - 100) | LZMA_PRESET_EXTREME;
-	else
-		preset = c->compression_level;
+		next = NULL;
+		if (p) {
+			next = p + 1;
+		} else {
+			p = token + strlen(token);
+			next = p;
+		}
 
-	if (lzma_lzma_preset(&ctx->opt, preset))
-		return -EINVAL;
-	ctx->opt.dict_size = c->dict_size;
+		if (!strncmp(token, "lc=", sizeof("lc=") - 1)) {
+			key = &ctx->opt.lc;
+			rhs = token + sizeof("lc=") - 1;
+		} else if (!strncmp(token, "lp=", sizeof("lp=") - 1)) {
+			key = &ctx->opt.lp;
+			rhs = token + sizeof("lp=") - 1;
+		} else if (!strncmp(token, "pb=", sizeof("pb=") - 1)) {
+			key = &ctx->opt.pb;
+			rhs = token + sizeof("pb=") - 1;
+		} else {
+			erofs_err("unknown extra options %s", extraopts);
+			return -EINVAL;
+		}
 
-	c->private_data = ctx;
+		val = strtoul(rhs, &endptr, 0);
+		if (val == ULONG_MAX || endptr != p) {
+			erofs_err("invalid option %.*s", p - token, token);
+			return -EINVAL;
+		}
+		*key = val;
+	}
 	return 0;
 }
 
@@ -118,10 +164,11 @@ const struct erofs_compressor erofs_compressor_lzma = {
 	.default_level = LZMA_PRESET_DEFAULT,
 	.best_level = 109,
 	.max_dictsize = Z_EROFS_LZMA_MAX_DICT_SIZE,
-	.init = erofs_compressor_liblzma_init,
+	.preinit = erofs_compressor_liblzma_preinit,
 	.exit = erofs_compressor_liblzma_exit,
 	.setlevel = erofs_compressor_liblzma_setlevel,
 	.setdictsize = erofs_compressor_liblzma_setdictsize,
+	.setextraopts = erofs_compressor_liblzma_setextraopts,
 	.compress_destsize = erofs_liblzma_compress_destsize,
 };
 #endif
