@@ -14,6 +14,23 @@ struct erofs_libzstd_context {
 	unsigned int fitblk_bufsiz;
 };
 
+static int libzstd_compress(const struct erofs_compress *c,
+			    const void *src, unsigned int srcsize,
+			    void *dst, unsigned int dstcapacity)
+{
+	struct erofs_libzstd_context *ctx = c->private_data;
+	size_t csize;
+
+	csize = ZSTD_compress2(ctx->cctx, dst, dstcapacity, src, srcsize);
+	if (ZSTD_isError(csize)) {
+		if (ZSTD_getErrorCode(csize) == ZSTD_error_dstSize_tooSmall)
+			return -ENOSPC;
+		erofs_err("Zstd compress failed: %s", ZSTD_getErrorName(csize));
+		return -EFAULT;
+	}
+	return csize;
+}
+
 static int libzstd_compress_destsize(const struct erofs_compress *c,
 				     const void *src, unsigned int *srcsize,
 				     void *dst, unsigned int dstsize)
@@ -98,14 +115,14 @@ static int erofs_compressor_libzstd_setlevel(struct erofs_compress *c,
 }
 
 static int erofs_compressor_libzstd_setdictsize(struct erofs_compress *c,
-						u32 dict_size)
+						u32 dict_size, u32 pclustersize_max)
 {
 	if (!dict_size) {
 		if (erofs_compressor_libzstd.default_dictsize) {
 			dict_size = erofs_compressor_libzstd.default_dictsize;
 		} else {
 			dict_size = min_t(u32, Z_EROFS_ZSTD_MAX_DICT_SIZE,
-					  cfg.c_mkfs_pclustersize_max << 3);
+					  pclustersize_max << 3);
 			dict_size = 1 << ilog2(dict_size);
 		}
 	}
@@ -158,8 +175,7 @@ static int compressor_libzstd_init(struct erofs_compress *c)
 
 	if (!erofs_atomic_test_and_set(&__warnonce)) {
 		erofs_warn("EXPERIMENTAL libzstd compressor in use. Note that `fitblk` isn't supported by upstream zstd for now.");
-		erofs_warn("Therefore it will takes more time in order to get the optimal result.");
-		erofs_info("You could clarify further needs in zstd repository <https://github.com/facebook/zstd/issues> for reference too.");
+		erofs_warn("If unaligned compression isn't used (without -E48bit), it will take more time in order to get the optimal result.");
 	}
 	return 0;
 out_err:
@@ -176,5 +192,6 @@ const struct erofs_compressor erofs_compressor_libzstd = {
 	.exit = compressor_libzstd_exit,
 	.setlevel = erofs_compressor_libzstd_setlevel,
 	.setdictsize = erofs_compressor_libzstd_setdictsize,
+	.compress = libzstd_compress,
 	.compress_destsize = libzstd_compress_destsize,
 };
