@@ -14,12 +14,13 @@
 #include <sys/mman.h>
 #include "erofs/err.h"
 #include "erofs/inode.h"
-#include "erofs/compress.h"
 #include "erofs/print.h"
 #include "erofs/internal.h"
-#include "erofs/fragments.h"
 #include "erofs/bitops.h"
 #include "erofs/lock.h"
+#include "erofs/importer.h"
+#include "liberofs_compress.h"
+#include "liberofs_fragments.h"
 #include "liberofs_private.h"
 #ifdef HAVE_SYS_SENDFILE_H
 #include <sys/sendfile.h>
@@ -52,7 +53,8 @@ struct erofs_packed_inode {
 
 const char *erofs_frags_packedname = "packed_file";
 
-u32 z_erofs_fragments_tofh(struct erofs_inode *inode, int fd, erofs_off_t fpos)
+u32 z_erofs_fragments_tofh(struct erofs_inode *inode,
+			   struct erofs_vfile *vf, erofs_off_t fpos)
 {
 	u8 data_to_hash[EROFS_TOF_HASHLEN];
 	u32 hash;
@@ -61,10 +63,10 @@ u32 z_erofs_fragments_tofh(struct erofs_inode *inode, int fd, erofs_off_t fpos)
 	if (inode->i_size <= EROFS_TOF_HASHLEN)
 		return ~0U;
 
-	ret = pread(fd, data_to_hash, EROFS_TOF_HASHLEN,
-		    fpos + inode->i_size - EROFS_TOF_HASHLEN);
+	ret = erofs_io_pread(vf, data_to_hash, EROFS_TOF_HASHLEN,
+			     fpos + inode->i_size - EROFS_TOF_HASHLEN);
 	if (ret < 0)
-		return -errno;
+		return ret;
 	if (ret != EROFS_TOF_HASHLEN) {
 		DBG_BUGON(1);
 		return -EIO;
@@ -75,7 +77,9 @@ u32 z_erofs_fragments_tofh(struct erofs_inode *inode, int fd, erofs_off_t fpos)
 
 static erofs_off_t erofs_fragment_longmatch(struct erofs_inode *inode,
 					    struct erofs_fragmentitem *fi,
-					    erofs_off_t matched, int fd)
+					    erofs_off_t matched,
+					    struct erofs_vfile *vf,
+					    erofs_off_t fpos)
 {
 	struct erofs_packed_inode *epi = inode->sbi->packedinode;
 	erofs_off_t total = min_t(erofs_off_t, fi->length, inode->i_size);
@@ -98,7 +102,8 @@ static erofs_off_t erofs_fragment_longmatch(struct erofs_inode *inode,
 			return matched;
 		}
 		sz = min_t(u64, total - matched, sizeof(buf[0]));
-		if (pread(fd, buf[0], sz, inode->i_size - matched - sz) != sz)
+		if (erofs_io_pread(vf, buf[0], sz,
+				   fpos + inode->i_size - matched - sz) != sz)
 			break;
 
 		if (!inmem) {
@@ -115,7 +120,8 @@ static erofs_off_t erofs_fragment_longmatch(struct erofs_inode *inode,
 	return matched;
 }
 
-int erofs_fragment_findmatch(struct erofs_inode *inode, int fd, u32 tofh)
+int erofs_fragment_findmatch(struct erofs_inode *inode,
+			     struct erofs_vfile *vf, erofs_off_t fpos, u32 tofh)
 {
 	struct erofs_packed_inode *epi = inode->sbi->packedinode;
 	struct erofs_fragmentitem *cur, *fi = NULL;
@@ -135,7 +141,7 @@ int erofs_fragment_findmatch(struct erofs_inode *inode, int fd, u32 tofh)
 	if (!data)
 		return -ENOMEM;
 
-	ret = pread(fd, data, s1, inode->i_size - s1);
+	ret = erofs_io_pread(vf, data, s1, fpos + inode->i_size - s1);
 	if (ret != s1) {
 		free(data);
 		return -errno;
@@ -166,7 +172,7 @@ int erofs_fragment_findmatch(struct erofs_inode *inode, int fd, u32 tofh)
 		i += EROFS_TOF_HASHLEN;
 		if (i >= s1) {		/* full short match */
 			DBG_BUGON(i > s1);
-			i = erofs_fragment_longmatch(inode, cur, s1, fd);
+			i = erofs_fragment_longmatch(inode, cur, s1, vf, fpos);
 		}
 
 		if (i <= deduped)
@@ -228,7 +234,8 @@ int erofs_fragment_pack(struct erofs_inode *inode, void *data,
 	return 0;
 }
 
-int erofs_pack_file_from_fd(struct erofs_inode *inode, int fd, u32 tofh)
+int erofs_pack_file_from_fd(struct erofs_inode *inode,
+			    struct erofs_vfile *vf, erofs_off_t fpos, u32 tofh)
 {
 	struct erofs_packed_inode *epi = inode->sbi->packedinode;
 	s64 offset, rc, sz;
@@ -242,33 +249,29 @@ int erofs_pack_file_from_fd(struct erofs_inode *inode, int fd, u32 tofh)
 	if (offset < 0)
 		return -errno;
 
-	memblock = mmap(NULL, inode->i_size, PROT_READ, MAP_SHARED, fd, 0);
+	if (vf->ops)
+		memblock = NULL;
+	else
+		memblock = mmap(NULL, inode->i_size, PROT_READ,
+				MAP_SHARED, vf->fd, fpos);
 	if (memblock == MAP_FAILED || !memblock) {
 		erofs_off_t remaining = inode->i_size;
-		struct erofs_vfile vin = { .fd = fd };
+		struct erofs_vfile vout = { .fd = epi->fd };
+		bool noseek = vf->ops && !vf->ops->pread;
+		off_t pos = fpos;
 
-#if defined(HAVE_SYS_SENDFILE_H) && defined(HAVE_SENDFILE)
 		do {
 			sz = min_t(u64, remaining, UINT_MAX);
-			rc = sendfile(epi->fd, fd, NULL, sz);
+			rc = erofs_io_sendfile(&vout, vf,
+					       noseek ? NULL : &pos, sz);
 			if (rc <= 0)
 				break;
 			remaining -= rc;
 		} while (remaining);
-#endif
-		while (remaining) {
-			char buf[32768];
 
-			sz = min_t(u64, remaining, sizeof(buf));
-			rc = erofs_io_read(&vin, buf, sz);
-			if (rc < 0)
-				goto out;
-			if (rc > 0) {
-				rc = write(epi->fd, buf, rc);
-				if (rc < 0)
-					goto out;
-			}
-			remaining -= rc;
+		if (remaining && rc >= 0) {
+			rc = -EIO;
+			goto out;
 		}
 
 		sz = min_t(u64, inode->i_size, EROFS_FRAGMENT_INMEM_SZ_MAX);
@@ -287,12 +290,6 @@ int erofs_pack_file_from_fd(struct erofs_inode *inode, int fd, u32 tofh)
 			}
 			goto out;
 		}
-
-		rc = lseek(fd, 0, SEEK_SET);
-		if (rc < 0) {
-			rc = -errno;
-			goto out;
-		}
 	} else {
 		rc = __erofs_io_write(epi->fd, memblock, inode->i_size);
 		if (rc != inode->i_size) {
@@ -370,17 +367,20 @@ int erofs_fragment_commit(struct erofs_inode *inode, u32 tofh)
 	return 0;
 }
 
-int erofs_flush_packed_inode(struct erofs_sb_info *sbi)
+int erofs_flush_packed_inode(struct erofs_importer *im)
 {
+	struct erofs_sb_info *sbi = im->sbi;
 	struct erofs_packed_inode *epi = sbi->packedinode;
 	struct erofs_inode *inode;
 
 	if (!epi || !erofs_sb_has_fragments(sbi))
-		return -EINVAL;
+		return 0;
 
 	if (lseek(epi->fd, 0, SEEK_CUR) <= 0)
 		return 0;
-	inode = erofs_mkfs_build_special_from_fd(sbi, epi->fd,
+
+	erofs_update_progressinfo("Processing packed data ...");
+	inode = erofs_mkfs_build_special_from_fd(im, epi->fd,
 						 EROFS_PACKED_INODE);
 	sbi->packed_nid = erofs_lookupnid(inode);
 	erofs_iput(inode);
@@ -510,6 +510,7 @@ static void *erofs_packedfile_preload(struct erofs_inode *pi,
 	struct erofs_sb_info *sbi = pi->sbi;
 	struct erofs_packed_inode *epi = sbi->packedinode;
 	unsigned int bsz = erofs_blksiz(sbi);
+	struct erofs_vfile vf;
 	char *buffer;
 	erofs_off_t pos, end;
 	ssize_t err;
@@ -529,13 +530,17 @@ static void *erofs_packedfile_preload(struct erofs_inode *pi,
 	else
 		DBG_BUGON(map->m_la > pos);
 
+	err = erofs_iopen(&vf, pi);
+	if (err)
+		return ERR_PTR(err);
+
 	map->m_llen = end - map->m_la;
 	DBG_BUGON(!map->m_llen);
 	buffer = malloc(map->m_llen);
 	if (!buffer)
 		return ERR_PTR(-ENOMEM);
 
-	err = erofs_pread(pi, buffer, map->m_llen, map->m_la);
+	err = erofs_pread(&vf, buffer, map->m_llen, map->m_la);
 	if (err)
 		goto err_out;
 
@@ -569,18 +574,20 @@ int erofs_packedfile_read(struct erofs_sb_info *sbi,
 	struct erofs_inode pi = {
 		.sbi = sbi,
 	};
-	struct erofs_map_blocks map = {
-		.index = UINT_MAX,
-	};
+	struct erofs_map_blocks map = { .buf = __EROFS_BUF_INITIALIZER };
 	unsigned int bsz = erofs_blksiz(sbi);
 	erofs_off_t end = pos + len;
+	struct erofs_vfile vf;
 	char *buffer = NULL;
 	int err;
 
 	if (!epi) {
 		err = erofs_load_packedinode_from_disk(&pi);
-		if (!err)
-			err = erofs_pread(&pi, buf, len, pos);
+		if (!err) {
+			err = erofs_iopen(&vf, &pi);
+			if (!err)
+				err = erofs_pread(&vf, buf, len, pos);
+		}
 		return err;
 	}
 
@@ -634,8 +641,11 @@ int erofs_packedfile_read(struct erofs_sb_info *sbi,
 			} else {
 fallback:
 				err = erofs_load_packedinode_from_disk(&pi);
+				if (err)
+					break;
+				err = erofs_iopen(&vf, &pi);
 				if (!err)
-					err = erofs_pread(&pi, buf, len, pos);
+					err = erofs_pread(&vf, buf, len, pos);
 				if (err)
 					break;
 			}
