@@ -147,10 +147,29 @@ int erofs_io_fsync(struct erofs_vfile *vf)
 	return 0;
 }
 
-ssize_t erofs_io_fallocate(struct erofs_vfile *vf, u64 offset,
-			   size_t len, bool zeroout)
+static const char erofs_zeroed[EROFS_MAX_BLOCK_SIZE];
+
+int __erofs_0write(int fd, size_t len)
+{
+	int err = 0;
+
+	while (len) {
+		u32 count = min_t(u64, sizeof(erofs_zeroed), len);
+
+		err = write(fd, erofs_zeroed, count);
+		if (err <= 0) {
+			if (err < 0)
+				err = -errno;
+			break;
+		}
+		len -= err;
+	}
+	return err < 0 ? err : len;
+}
+
+int erofs_io_fallocate(struct erofs_vfile *vf, u64 offset,
+		       size_t len, bool zeroout)
 {
-	static const char zero[EROFS_MAX_BLOCK_SIZE] = {0};
 	ssize_t ret;
 
 	if (__erofs_unlikely(cfg.c_dry_run))
@@ -164,14 +183,15 @@ ssize_t erofs_io_fallocate(struct erofs_vfile *vf, u64 offset,
 		    FALLOC_FL_KEEP_SIZE, offset + vf->offset, len) >= 0)
 		return 0;
 #endif
-	while (len > EROFS_MAX_BLOCK_SIZE) {
-		ret = erofs_io_pwrite(vf, zero, offset, EROFS_MAX_BLOCK_SIZE);
+	while (len > sizeof(erofs_zeroed)) {
+		ret = erofs_io_pwrite(vf, erofs_zeroed, offset,
+				      sizeof(erofs_zeroed));
 		if (ret < 0)
-			return ret;
+			return (int)ret;
 		len -= ret;
 		offset += ret;
 	}
-	return erofs_io_pwrite(vf, zero, offset, len) == len ? 0 : -EIO;
+	return erofs_io_pwrite(vf, erofs_zeroed, offset, len) == len ? 0 : -EIO;
 }
 
 int erofs_io_ftruncate(struct erofs_vfile *vf, u64 length)
@@ -196,7 +216,7 @@ int erofs_io_ftruncate(struct erofs_vfile *vf, u64 length)
 	return ftruncate(vf->fd, length);
 }
 
-ssize_t erofs_io_pread(struct erofs_vfile *vf, void *buf, u64 pos, size_t len)
+ssize_t erofs_io_pread(struct erofs_vfile *vf, void *buf, size_t len, u64 pos)
 {
 	ssize_t ret, read = 0;
 
@@ -204,7 +224,7 @@ ssize_t erofs_io_pread(struct erofs_vfile *vf, void *buf, u64 pos, size_t len)
 		return 0;
 
 	if (vf->ops)
-		return vf->ops->pread(vf, buf, pos, len);
+		return vf->ops->pread(vf, buf, len, pos);
 
 	pos += vf->offset;
 	do {
@@ -360,8 +380,7 @@ out:
 
 void erofs_dev_close(struct erofs_sb_info *sbi)
 {
-	if (!sbi->bdev.ops)
-		close(sbi->bdev.fd);
+	erofs_io_close(&sbi->bdev);
 	free(sbi->devname);
 	sbi->devname = NULL;
 	sbi->bdev.fd = -1;
@@ -403,9 +422,9 @@ ssize_t erofs_dev_read(struct erofs_sb_info *sbi, int device_id,
 		}
 		read = erofs_io_pread(&((struct erofs_vfile) {
 				.fd = sbi->blobfd[device_id - 1],
-			}), buf, offset, len);
+			}), buf, len, offset);
 	} else {
-		read = erofs_io_pread(&sbi->bdev, buf, offset, len);
+		read = erofs_io_pread(&sbi->bdev, buf, len, offset);
 	}
 
 	if (read < 0)
@@ -539,16 +558,78 @@ ssize_t erofs_io_read(struct erofs_vfile *vf, void *buf, size_t bytes)
         return i;
 }
 
+ssize_t erofs_io_write(struct erofs_vfile *vf, void *buf, size_t len)
+{
+	if (vf->ops)
+		return vf->ops->write(vf, buf, len);
+	return __erofs_io_write(vf->fd, buf, len);
+}
+
 #ifdef HAVE_SYS_SENDFILE_H
 #include <sys/sendfile.h>
 #endif
 
 off_t erofs_io_lseek(struct erofs_vfile *vf, u64 offset, int whence)
 {
+	off_t ret;
+
 	if (vf->ops)
 		return vf->ops->lseek(vf, offset, whence);
 
-	return lseek(vf->fd, offset, whence);
+	ret = lseek(vf->fd, offset, whence);
+	return ret < 0 ? -errno : ret;
+}
+
+ssize_t erofs_io_sendfile(struct erofs_vfile *vout, struct erofs_vfile *vin,
+			  off_t *pos, size_t count)
+{
+	ssize_t read, written;
+	size_t rem = count;
+
+	if (vin->ops || vout->ops) {
+		if (vin->ops && vin->ops->sendfile)
+			return vin->ops->sendfile(vout, vin, pos, count);
+		if (vout->ops && vout->ops->sendfile)
+			return vout->ops->sendfile(vout, vin, pos, count);
+		written = 0;
+	}
+#if defined(HAVE_SYS_SENDFILE_H) && defined(HAVE_SENDFILE)
+	else do {
+		written = sendfile(vout->fd, vin->fd, pos, rem);
+		if (written <= 0) {
+			if (written < 0) {
+				written = -errno;
+				if (written == -EOVERFLOW && pos)
+					written = 0;
+			}
+			break;
+		}
+		rem -= written;
+	} while (written);
+#endif
+	while (rem) {
+		char buf[max(EROFS_MAX_BLOCK_SIZE, 32768)];
+
+		read = min_t(u64, rem, sizeof(buf));
+		if (pos)
+			read = erofs_io_pread(vin, buf, read, *pos);
+		else
+			read = erofs_io_read(vin, buf, read);
+		if (read <= 0) {
+			written = read;
+			break;
+		}
+		rem -= read;
+		if (pos)
+			*pos += read;
+		do {
+			written = erofs_io_write(vout, buf, read);
+			if (written < 0)
+				return written;
+			read -= written;
+		} while (read);
+	}
+	return written < 0 ? written : count - rem;
 }
 
 int erofs_io_xcopy(struct erofs_vfile *vout, off_t pos,
@@ -596,3 +677,13 @@ int erofs_io_xcopy(struct erofs_vfile *vout, off_t pos,
 	} while (len);
 	return 0;
 }
+
+void erofs_io_close(struct erofs_vfile *vf)
+{
+	if (vf->ops) {
+		vf->ops->close(vf);
+		return;
+	}
+	close(vf->fd);
+	vf->fd = -1;
+}
