@@ -10,11 +10,11 @@
 #endif
 #include "erofs/print.h"
 #include "erofs/inode.h"
-#include "erofs/rebuild.h"
 #include "erofs/dir.h"
 #include "erofs/xattr.h"
 #include "erofs/blobchunk.h"
 #include "erofs/internal.h"
+#include "liberofs_rebuild.h"
 #include "liberofs_uuid.h"
 
 #ifdef HAVE_LINUX_AUFS_TYPE_H
@@ -59,7 +59,7 @@ static struct erofs_dentry *erofs_rebuild_mkdir(struct erofs_inode *dir,
 	inode->i_mtime = dir->i_mtime;
 	inode->i_mtime_nsec = dir->i_mtime_nsec;
 	inode->dev = dir->dev;
-	erofs_init_empty_dir(inode);
+	inode->i_nlink = 2;
 
 	d = erofs_d_alloc(dir, s);
 	if (IS_ERR(d)) {
@@ -186,7 +186,7 @@ static int erofs_rebuild_write_blob_index(struct erofs_sb_info *dst_sb,
 
 	unit = sizeof(struct erofs_inode_chunk_index);
 	inode->extent_isize = count * unit;
-	idx = malloc(max(sizeof(*idx), sizeof(void *)));
+	idx = calloc(count, max(sizeof(*idx), sizeof(void *)));
 	if (!idx)
 		return -ENOMEM;
 	inode->chunkindexes = idx;
@@ -194,7 +194,7 @@ static int erofs_rebuild_write_blob_index(struct erofs_sb_info *dst_sb,
 	for (i = 0; i < count; i++) {
 		struct erofs_blobchunk *chunk;
 		struct erofs_map_blocks map = {
-			.index = UINT_MAX,
+			.buf = __EROFS_BUF_INITIALIZER,
 		};
 
 		map.m_la = i << chunkbits;
@@ -231,7 +231,7 @@ static int erofs_rebuild_update_inode(struct erofs_sb_info *dst_sb,
 	case S_IFCHR:
 		if (erofs_inode_is_whiteout(inode))
 			inode->i_parent->whiteouts = true;
-		/* fallthrough */
+		__erofs_fallthrough;
 	case S_IFBLK:
 	case S_IFIFO:
 	case S_IFSOCK:
@@ -241,18 +241,24 @@ static int erofs_rebuild_update_inode(struct erofs_sb_info *dst_sb,
 		inode->u.i_rdev = erofs_new_encode_dev(inode->u.i_rdev);
 		break;
 	case S_IFDIR:
-		err = erofs_init_empty_dir(inode);
+		inode->i_nlink = 2;
 		break;
-	case S_IFLNK:
+	case S_IFLNK: {
+		struct erofs_vfile vf;
+
 		inode->i_link = malloc(inode->i_size + 1);
 		if (!inode->i_link)
 			return -ENOMEM;
-		err = erofs_pread(inode, inode->i_link, inode->i_size, 0);
+		err = erofs_iopen(&vf, inode);
+		if (err)
+			return err;
+		err = erofs_pread(&vf, inode->i_link, inode->i_size, 0);
 		erofs_dbg("\tsymlink: %s -> %s", inode->i_srcpath, inode->i_link);
 		break;
+	}
 	case S_IFREG:
 		if (!inode->i_size) {
-			inode->u.i_blkaddr = NULL_ADDR;
+			inode->u.i_blkaddr = EROFS_NULL_ADDR;
 			break;
 		}
 		if (datamode == EROFS_REBUILD_DATA_BLOB_INDEX)
@@ -268,15 +274,18 @@ static int erofs_rebuild_update_inode(struct erofs_sb_info *dst_sb,
 	return err;
 }
 
-/*
- * @mergedir: parent directory in the merged tree
- * @ctx.dir:  parent directory when itering erofs_iterate_dir()
- * @datamode: indicate how to import inode data
- */
 struct erofs_rebuild_dir_context {
+	/* @ctx.dir:  parent directory when itering erofs_iterate_dir() */
 	struct erofs_dir_context ctx;
-	struct erofs_inode *mergedir;
-	enum erofs_rebuild_datamode datamode;
+	struct erofs_inode *mergedir;	/* parent directory in the merged tree */
+	union {
+		/* indicate how to import inode data */
+		enum erofs_rebuild_datamode datamode;
+		struct {
+			u64 *nr_subdirs;
+			unsigned int *i_nlink;
+		};
+	};
 };
 
 static int erofs_rebuild_dirent_iter(struct erofs_dir_context *ctx)
@@ -403,7 +412,7 @@ static int erofs_rebuild_dirent_iter(struct erofs_dir_context *ctx)
 
 	/* reset sbi, nid after subdirs are all loaded for the final dump */
 	inode->sbi = &g_sbi;
-	inode->nid = 0;
+	inode->nid = EROFS_NID_UNALLOCATED;
 out:
 	free(path);
 	return ret;
@@ -422,6 +431,7 @@ int erofs_rebuild_load_tree(struct erofs_inode *root, struct erofs_sb_info *sbi,
 		erofs_uuid_unparse_lower(sbi->uuid, uuid_str);
 		fsid = uuid_str;
 	}
+
 	ret = erofs_read_superblock(sbi);
 	if (ret) {
 		erofs_err("failed to read superblock of %s", fsid);
@@ -451,8 +461,8 @@ int erofs_rebuild_load_tree(struct erofs_inode *root, struct erofs_sb_info *sbi,
 static int erofs_rebuild_basedir_dirent_iter(struct erofs_dir_context *ctx)
 {
 	struct erofs_rebuild_dir_context *rctx = (void *)ctx;
-	struct erofs_inode *dir = ctx->dir;
 	struct erofs_inode *mergedir = rctx->mergedir;
+	struct erofs_inode *dir = ctx->dir;
 	struct erofs_dentry *d;
 	char *dname;
 	bool dumb;
@@ -474,9 +484,18 @@ static int erofs_rebuild_basedir_dirent_iter(struct erofs_dir_context *ctx)
 	if (d->type == EROFS_FT_UNKNOWN) {
 		d->nid = ctx->de_nid;
 		d->type = ctx->de_ftype;
-		d->validnid = true;
+		d->flags |= EROFS_DENTRY_FLAG_VALIDNID;
+		if (d->type == EROFS_FT_DIR)
+			d->flags |= EROFS_DENTRY_FLAG_FIXUP_PNID;
 		if (!mergedir->whiteouts && erofs_dentry_is_wht(dir->sbi, d))
 			mergedir->whiteouts = true;
+		*rctx->i_nlink += (ctx->de_ftype == EROFS_FT_DIR);
+		++*rctx->nr_subdirs;
+	} else if (__erofs_unlikely(d->flags & EROFS_DENTRY_FLAG_VALIDNID)) {
+		/* The base image appears to be corrupted */
+		DBG_BUGON(1);
+		ret = -EFSCORRUPTED;
+		goto out;
 	} else {
 		struct erofs_inode *inode = d->inode;
 
@@ -496,7 +515,8 @@ out:
 	return ret;
 }
 
-int erofs_rebuild_load_basedir(struct erofs_inode *dir)
+int erofs_rebuild_load_basedir(struct erofs_inode *dir, u64 *nr_subdirs,
+			       unsigned int *i_nlink)
 {
 	struct erofs_inode fakeinode = {
 		.sbi = dir->sbi,
@@ -528,6 +548,8 @@ int erofs_rebuild_load_basedir(struct erofs_inode *dir)
 		.ctx.dir = &fakeinode,
 		.ctx.cb = erofs_rebuild_basedir_dirent_iter,
 		.mergedir = dir,
+		.nr_subdirs = nr_subdirs,
+		.i_nlink = i_nlink,
 	};
 	return erofs_iterate_dir(&ctx.ctx, false);
 }
