@@ -6,7 +6,9 @@
 #include <stdlib.h>
 #include "erofs/print.h"
 #include "erofs/xattr.h"
-#include "erofs/cache.h"
+#include "liberofs_cache.h"
+#include "liberofs_compress.h"
+#include "liberofs_metabox.h"
 
 static bool check_layout_compatibility(struct erofs_sb_info *sbi,
 				       struct erofs_super_block *dsb)
@@ -63,8 +65,8 @@ static int erofs_init_devices(struct erofs_sb_info *sbi,
 			return ret;
 		}
 
-		sbi->devs[i].mapped_blkaddr = le32_to_cpu(dis.mapped_blkaddr);
-		sbi->devs[i].blocks = le32_to_cpu(dis.blocks);
+		sbi->devs[i].uniaddr = le32_to_cpu(dis.uniaddr_lo);
+		sbi->devs[i].blocks = le32_to_cpu(dis.blocks_lo);
 		memcpy(sbi->devs[i].tag, dis.tag, sizeof(dis.tag));
 		sbi->total_blocks += sbi->devs[i].blocks;
 		pos += EROFS_DEVT_SLOT_SIZE;
@@ -78,10 +80,14 @@ int erofs_read_superblock(struct erofs_sb_info *sbi)
 	struct erofs_super_block *dsb;
 	int read, ret;
 
-	read = erofs_io_pread(&sbi->bdev, data, 0, EROFS_MAX_BLOCK_SIZE);
-	if (read < EROFS_SUPER_END) {
+	if (sbi->sb_valid)
+		return 0;
+
+	read = erofs_io_pread(&sbi->bdev, data, EROFS_MAX_BLOCK_SIZE, 0);
+	if (read < EROFS_SUPER_OFFSET + sizeof(*dsb)) {
 		ret = read < 0 ? read : -EIO;
-		erofs_err("cannot read erofs superblock: %d", ret);
+		erofs_err("cannot read erofs superblock: %s",
+			  erofs_strerror(ret));
 		return ret;
 	}
 	dsb = (struct erofs_super_block *)(data + EROFS_SUPER_OFFSET);
@@ -109,22 +115,46 @@ int erofs_read_superblock(struct erofs_sb_info *sbi)
 		erofs_err("invalid sb_extslots %u", dsb->sb_extslots);
 		return -EINVAL;
 	}
-	sbi->primarydevice_blocks = le32_to_cpu(dsb->blocks);
+	sbi->primarydevice_blocks = le32_to_cpu(dsb->blocks_lo);
 	sbi->meta_blkaddr = le32_to_cpu(dsb->meta_blkaddr);
 	sbi->xattr_blkaddr = le32_to_cpu(dsb->xattr_blkaddr);
 	sbi->xattr_prefix_start = le32_to_cpu(dsb->xattr_prefix_start);
 	sbi->xattr_prefix_count = dsb->xattr_prefix_count;
-	sbi->islotbits = EROFS_ISLOTBITS;
-	sbi->root_nid = le16_to_cpu(dsb->root_nid);
+	if (erofs_sb_has_48bit(sbi) && dsb->rootnid_8b) {
+		sbi->root_nid = le64_to_cpu(dsb->rootnid_8b);
+		sbi->primarydevice_blocks = (sbi->primarydevice_blocks << 32) |
+				le16_to_cpu(dsb->rb.blocks_hi);
+	} else {
+		sbi->root_nid = le16_to_cpu(dsb->rb.rootnid_2b);
+	}
 	sbi->packed_nid = le64_to_cpu(dsb->packed_nid);
+	if (erofs_sb_has_metabox(sbi)) {
+		if (sbi->sb_size <= offsetof(struct erofs_super_block,
+					     metabox_nid))
+			return -EFSCORRUPTED;
+		sbi->metabox_nid = le64_to_cpu(dsb->metabox_nid);
+		if (sbi->metabox_nid & BIT_ULL(EROFS_DIRENT_NID_METABOX_BIT))
+			return -EFSCORRUPTED;	/* self-loop detection */
+	} else {
+		sbi->metabox_nid = 0;
+	}
 	sbi->inos = le64_to_cpu(dsb->inos);
 	sbi->checksum = le32_to_cpu(dsb->checksum);
 
-	sbi->build_time = le64_to_cpu(dsb->build_time);
-	sbi->build_time_nsec = le32_to_cpu(dsb->build_time_nsec);
+	sbi->epoch = (s64)le64_to_cpu(dsb->epoch);
+	sbi->fixed_nsec = le32_to_cpu(dsb->fixed_nsec);
+	sbi->build_time = le32_to_cpu(dsb->build_time);
 
 	memcpy(&sbi->uuid, dsb->uuid, sizeof(dsb->uuid));
-
+	if (erofs_sb_has_ishare_xattrs(sbi)) {
+		if (dsb->ishare_xattr_prefix_id >= sbi->xattr_prefix_count) {
+			erofs_err("invalid ishare xattr prefix id %d",
+				  dsb->ishare_xattr_prefix_id);
+			return -EFSCORRUPTED;
+		}
+		sbi->ishare_xattr_prefix_id =
+			dsb->ishare_xattr_prefix_id | EROFS_XATTR_LONG_PREFIX;
+	}
 	ret = z_erofs_parse_cfgs(sbi, dsb);
 	if (ret)
 		return ret;
@@ -138,6 +168,13 @@ int erofs_read_superblock(struct erofs_sb_info *sbi)
 		free(sbi->devs);
 		sbi->devs = NULL;
 	}
+	sbi->sb_valid = !ret;
+	if (erofs_sb_has_48bit(sbi))
+		erofs_info("EXPERIMENTAL 48-bit layout support in use. Use at your own risk!");
+	if (erofs_sb_has_metabox(sbi)) {
+		erofs_info("EXPERIMENTAL metadata compression support in use. Use at your own risk!");
+		erofs_info("No in-memory cache for metadata compression: userspace parser for metabox remains slow.");
+	}
 	return ret;
 }
 
@@ -152,22 +189,26 @@ void erofs_put_super(struct erofs_sb_info *sbi)
 		free(sbi->devs);
 		sbi->devs = NULL;
 	}
-	erofs_xattr_prefixes_cleanup(sbi);
 	if (sbi->bmgr) {
 		erofs_buffer_exit(sbi->bmgr);
 		sbi->bmgr = NULL;
 	}
+	z_erofs_compress_exit(sbi);
+	erofs_xattr_exit(sbi);
+	sbi->sb_valid = false;
 }
 
-int erofs_writesb(struct erofs_sb_info *sbi, struct erofs_buffer_head *sb_bh)
+int erofs_writesb(struct erofs_sb_info *sbi)
 {
+	struct erofs_buffer_head *sb_bh = sbi->bh_sb;
 	struct erofs_super_block sb = {
 		.magic     = cpu_to_le32(EROFS_SUPER_MAGIC_V1),
 		.blkszbits = sbi->blkszbits,
-		.root_nid  = cpu_to_le16(sbi->root_nid),
+		.rb.rootnid_2b  = cpu_to_le16(sbi->root_nid),
 		.inos      = cpu_to_le64(sbi->inos),
+		.epoch     = cpu_to_le64(sbi->epoch),
 		.build_time = cpu_to_le64(sbi->build_time),
-		.build_time_nsec = cpu_to_le32(sbi->build_time_nsec),
+		.fixed_nsec = cpu_to_le32(sbi->fixed_nsec),
 		.meta_blkaddr  = cpu_to_le32(sbi->meta_blkaddr),
 		.xattr_blkaddr = cpu_to_le32(sbi->xattr_blkaddr),
 		.xattr_prefix_count = sbi->xattr_prefix_count,
@@ -178,12 +219,18 @@ int erofs_writesb(struct erofs_sb_info *sbi, struct erofs_buffer_head *sb_bh)
 		.extra_devices = cpu_to_le16(sbi->extra_devices),
 		.devt_slotoff = cpu_to_le16(sbi->devt_slotoff),
 		.packed_nid = cpu_to_le64(sbi->packed_nid),
+		.ishare_xattr_prefix_id = sbi->ishare_xattr_prefix_id &
+			EROFS_XATTR_LONG_PREFIX_MASK,
 	};
-	const u32 sb_blksize = round_up(EROFS_SUPER_END, erofs_blksiz(sbi));
 	char *buf;
 	int ret;
 
-	sb.blocks       = cpu_to_le32(sbi->primarydevice_blocks);
+	sb.blocks_lo	= cpu_to_le32(sbi->primarydevice_blocks);
+	if (sbi->primarydevice_blocks > UINT32_MAX ||
+	    sbi->root_nid > UINT16_MAX) {
+		sb.rb.blocks_hi = cpu_to_le16(sbi->primarydevice_blocks >> 32);
+		sb.rootnid_8b = cpu_to_le64(sbi->root_nid);
+	}
 	memcpy(sb.uuid, sbi->uuid, sizeof(sb.uuid));
 	memcpy(sb.volume_name, sbi->volume_name, sizeof(sb.volume_name));
 
@@ -192,16 +239,21 @@ int erofs_writesb(struct erofs_sb_info *sbi, struct erofs_buffer_head *sb_bh)
 	else
 		sb.u1.lz4_max_distance = cpu_to_le16(sbi->lz4.max_distance);
 
-	buf = calloc(sb_blksize, 1);
+	if (erofs_sb_has_metabox(sbi))
+		sb.metabox_nid = cpu_to_le64(sbi->metabox_nid);
+	sb.sb_extslots = (sbi->sb_size - 128) >> 4;
+
+	buf = calloc(round_up(EROFS_SUPER_OFFSET + sbi->sb_size,
+			      erofs_blksiz(sbi)), 1);
 	if (!buf) {
 		erofs_err("failed to allocate memory for sb: %s",
 			  erofs_strerror(-errno));
 		return -ENOMEM;
 	}
-	memcpy(buf + EROFS_SUPER_OFFSET, &sb, sizeof(sb));
+	memcpy(buf + EROFS_SUPER_OFFSET, &sb, sbi->sb_size);
 
 	ret = erofs_dev_write(sbi, buf, sb_bh ? erofs_btell(sb_bh, false) : 0,
-			      EROFS_SUPER_END);
+			      EROFS_SUPER_OFFSET + sbi->sb_size);
 	free(buf);
 	if (sb_bh)
 		erofs_bdrop(sb_bh, false);
@@ -210,9 +262,16 @@ int erofs_writesb(struct erofs_sb_info *sbi, struct erofs_buffer_head *sb_bh)
 
 struct erofs_buffer_head *erofs_reserve_sb(struct erofs_bufmgr *bmgr)
 {
+	struct erofs_sb_info *sbi = bmgr->sbi;
 	struct erofs_buffer_head *bh;
+	unsigned int sb_size = 128;
 	int err;
 
+	if (erofs_sb_has_metabox(sbi) &&
+	    sb_size <= offsetof(struct erofs_super_block, metabox_nid))
+		sb_size = offsetof(struct erofs_super_block, metabox_nid) + 8;
+	sbi->sb_size = round_up(sb_size, 16);
+
 	bh = erofs_balloc(bmgr, META, 0, 0);
 	if (IS_ERR(bh)) {
 		erofs_err("failed to allocate super: %s",
@@ -220,7 +279,7 @@ struct erofs_buffer_head *erofs_reserve_sb(struct erofs_bufmgr *bmgr)
 		return bh;
 	}
 	bh->op = &erofs_skip_write_bhops;
-	err = erofs_bh_balloon(bh, EROFS_SUPER_END);
+	err = erofs_bh_balloon(bh, EROFS_SUPER_OFFSET + sbi->sb_size);
 	if (err < 0) {
 		erofs_err("failed to balloon super: %s", erofs_strerror(err));
 		goto err_bdrop;
@@ -347,11 +406,14 @@ int erofs_write_device_table(struct erofs_sb_info *sbi)
 
 	if (!sbi->extra_devices)
 		goto out;
-	if (!bh)
+	if (!bh) {
+		if (erofs_sb_has_device_table(sbi))
+			return 0;
 		return -EINVAL;
+	}
 
 	pos = erofs_btell(bh, false);
-	if (pos == NULL_ADDR_UL) {
+	if (pos == EROFS_NULL_ADDR) {
 		DBG_BUGON(1);
 		return -EINVAL;
 	}
@@ -359,8 +421,8 @@ int erofs_write_device_table(struct erofs_sb_info *sbi)
 	i = 0;
 	do {
 		struct erofs_deviceslot dis = {
-			.mapped_blkaddr = cpu_to_le32(nblocks),
-			.blocks = cpu_to_le32(sbi->devs[i].blocks),
+			.uniaddr_lo = cpu_to_le32(nblocks),
+			.blocks_lo = cpu_to_le32(sbi->devs[i].blocks),
 		};
 
 		memcpy(dis.tag, sbi->devs[i].tag, sizeof(dis.tag));
@@ -378,3 +440,58 @@ out:
 	sbi->total_blocks = nblocks;
 	return 0;
 }
+
+int erofs_mkfs_format_fs(struct erofs_sb_info *sbi, unsigned int blkszbits,
+			 unsigned int dsunit, bool metazone)
+{
+	struct erofs_buffer_head *bh;
+	struct erofs_bufmgr *bmgr;
+
+	sbi->blkszbits = blkszbits;
+	bmgr = erofs_buffer_init(sbi, 0, NULL);
+	if (!bmgr)
+		return -ENOMEM;
+	sbi->bmgr = bmgr;
+	bmgr->dsunit = dsunit;
+	if (metazone)
+		sbi->metazone_startblk = EROFS_META_NEW_ADDR;
+	else
+		sbi->metazone_startblk = 0;
+	bh = erofs_reserve_sb(bmgr);
+	if (IS_ERR(bh))
+		return PTR_ERR(bh);
+	sbi->bh_sb = bh;
+	return 0;
+}
+
+int erofs_mkfs_load_fs(struct erofs_sb_info *sbi, unsigned int dsunit)
+{
+	union {
+		struct stat st;
+		erofs_blk_t startblk;
+	} u;
+	struct erofs_bufmgr *bmgr;
+	int err;
+
+	sbi->bh_sb = NULL;
+	erofs_warn("EXPERIMENTAL incremental build in use. Use at your own risk!");
+	err = erofs_read_superblock(sbi);
+	if (err) {
+		erofs_err("failed to read superblock of %s: %s", sbi->devname,
+			  erofs_strerror(err));
+		return err;
+	}
+
+	err = erofs_io_fstat(&sbi->bdev, &u.st);
+	if (!err && S_ISREG(u.st.st_mode))
+		u.startblk = DIV_ROUND_UP(u.st.st_size, erofs_blksiz(sbi));
+	else
+		u.startblk = sbi->primarydevice_blocks;
+
+	bmgr = erofs_buffer_init(sbi, u.startblk, NULL);
+	if (!bmgr)
+		return -ENOMEM;
+	sbi->bmgr = bmgr;
+	bmgr->dsunit = dsunit;
+	return 0;
+}
