@@ -4,17 +4,20 @@
 #include <string.h>
 #include <sys/stat.h>
 #include "erofs/print.h"
-#include "erofs/cache.h"
 #include "erofs/diskbuf.h"
 #include "erofs/inode.h"
 #include "erofs/list.h"
 #include "erofs/tar.h"
 #include "erofs/xattr.h"
 #include "erofs/blobchunk.h"
-#include "erofs/rebuild.h"
+#include "erofs/importer.h"
 #if defined(HAVE_ZLIB)
 #include <zlib.h>
 #endif
+#include "liberofs_base64.h"
+#include "liberofs_cache.h"
+#include "liberofs_gzran.h"
+#include "liberofs_rebuild.h"
 
 /* This file is a tape/volume header.  Ignore it on extraction.  */
 #define GNUTYPE_VOLHDR 'V'
@@ -63,8 +66,11 @@ void erofs_iostream_close(struct erofs_iostream *ios)
 		free(ios->lzma);
 #endif
 		return;
+	} else if (ios->decoder == EROFS_IOS_DECODER_GZRAN) {
+		erofs_gzran_builder_final(ios->gb);
+		return;
 	}
-	close(ios->vf.fd);
+	erofs_io_close(&ios->vf);
 }
 
 int erofs_iostream_open(struct erofs_iostream *ios, int fd, int decoder)
@@ -103,6 +109,14 @@ int erofs_iostream_open(struct erofs_iostream *ios, int fd, int decoder)
 #else
 		return -EOPNOTSUPP;
 #endif
+	} else if (decoder == EROFS_IOS_DECODER_GZRAN) {
+		ios->vf.fd = fd;
+		ios->feof = false;
+		ios->sz = 0;
+		ios->bufsize = EROFS_GZRAN_WINSIZE * 2;
+		ios->gb = erofs_gzran_builder_init(&ios->vf, 4194304);
+		if (IS_ERR(ios->gb))
+			return PTR_ERR(ios->gb);
 	} else {
 		ios->vf.fd = fd;
 		fsz = lseek(fd, 0, SEEK_END);
@@ -202,6 +216,14 @@ int erofs_iostream_read(struct erofs_iostream *ios, void **buf, u64 bytes)
 #else
 			return -EOPNOTSUPP;
 #endif
+		} else if (ios->decoder == EROFS_IOS_DECODER_GZRAN) {
+			ret = erofs_gzran_builder_read(ios->gb, ios->buffer + rabytes);
+			if (ret < 0)
+				return ret;
+			ios->tail += ret;
+			DBG_BUGON(ios->tail > ios->bufsize);
+			if (!ret)
+				ios->feof = true;
 		} else {
 			ret = erofs_io_read(&ios->vf, ios->buffer + rabytes,
 					    ios->bufsize - rabytes);
@@ -389,7 +411,7 @@ int tarerofs_apply_xattrs(struct erofs_inode *inode, struct list_head *xattrs)
 		item->kv[item->namelen] = '\0';
 		erofs_dbg("Recording xattr(%s)=\"%s\" (of %u bytes) to file %s",
 			  item->kv, v, vsz, inode->i_srcpath);
-		ret = erofs_setxattr(inode, item->kv, v, vsz);
+		ret = erofs_vfs_setxattr(inode, item->kv, v, vsz);
 		if (ret == -ENODATA)
 			erofs_err("Failed to set xattr(%s)=%s to file %s",
 				  item->kv, v, inode->i_srcpath);
@@ -399,46 +421,6 @@ int tarerofs_apply_xattrs(struct erofs_inode *inode, struct list_head *xattrs)
 	return 0;
 }
 
-static const char lookup_table[65] =
-	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
-
-static int base64_decode(const char *src, int len, u8 *dst)
-{
-	int i, bits = 0, ac = 0;
-	const char *p;
-	u8 *cp = dst;
-	bool padding = false;
-
-	if(len && !(len % 4)) {
-		/* Check for and ignore any end padding */
-		if (src[len - 2] == '=' && src[len - 1] == '=')
-			len -= 2;
-		else if (src[len - 1] == '=')
-			--len;
-		padding = true;
-	}
-
-	for (i = 0; i < len; i++) {
-		p = strchr(lookup_table, src[i]);
-		if (!p || !src[i])
-			return -2;
-		ac += (p - lookup_table) << bits;
-		bits += 6;
-		if (bits >= 8) {
-			*cp++ = ac & 0xff;
-			ac >>= 8;
-			bits -= 8;
-		}
-	}
-	if (ac) {
-		if (padding || ac > 0xff)
-			return -1;
-		else
-			*cp++ = ac & 0xff;
-	}
-	return cp - dst;
-}
-
 static int tohex(int c)
 {
 	if (c >= '0' && c <= '9')
@@ -593,8 +575,8 @@ int tarerofs_parse_pax_header(struct erofs_iostream *ios,
 				key = kv + sizeof("LIBARCHIVE.xattr.") - 1;
 				namelen = url_decode(key, value - key - 1);
 				--len; /* p[-1] == '\0' */
-				ret = base64_decode(value, len - (value - kv),
-						    (u8 *)value);
+				ret = erofs_base64_decode(value,
+						len - (value - kv), (u8 *)value);
 				if (ret < 0) {
 					ret = -EFSCORRUPTED;
 					goto out;
@@ -650,7 +632,7 @@ static int tarerofs_write_uncompressed_file(struct erofs_inode *inode,
 	inode->datalayout = EROFS_INODE_FLAT_PLAIN;
 	nblocks = DIV_ROUND_UP(inode->i_size, 1U << sbi->blkszbits);
 
-	ret = erofs_allocate_inode_bh_data(inode, nblocks);
+	ret = erofs_allocate_inode_bh_data(inode, nblocks, false);
 	if (ret)
 		return ret;
 
@@ -693,23 +675,26 @@ static int tarerofs_write_file_data(struct erofs_inode *inode,
 		nread = erofs_iostream_read(&tar->ios, &buf, j);
 		if (nread < 0)
 			break;
-		if (write(fd, buf, nread) != nread) {
+		if (pwrite(fd, buf, nread, off) != nread) {
 			nread = -EIO;
 			break;
 		}
 		j -= nread;
+		off += nread;
 	}
 	erofs_diskbuf_commit(inode->i_diskbuf, inode->i_size);
 	inode->datasource = EROFS_INODE_DATA_SOURCE_DISKBUF;
 	return 0;
 }
 
-int tarerofs_parse_tar(struct erofs_inode *root, struct erofs_tarfile *tar)
+int tarerofs_parse_tar(struct erofs_importer *im, struct erofs_tarfile *tar)
 {
-	char path[PATH_MAX];
 	struct erofs_pax_header eh = tar->global;
-	struct erofs_sb_info *sbi = root->sbi;
+	struct erofs_importer_params *params = im->params;
+	struct erofs_sb_info *sbi = im->sbi;
+	struct erofs_inode *root = im->root;
 	bool whout, opq, e = false;
+	char _path[PATH_MAX + 1], *path = _path + 1;
 	struct stat st;
 	mode_t mode;
 	erofs_off_t tar_offset, dataoff;
@@ -717,8 +702,8 @@ int tarerofs_parse_tar(struct erofs_inode *root, struct erofs_tarfile *tar)
 	struct tar_header *th;
 	struct erofs_dentry *d;
 	struct erofs_inode *inode;
-	unsigned int j, csum, cksum;
-	int ckksum, ret, rem;
+	unsigned int csum, cksum;
+	int ckksum, ret, rem, j;
 
 	root->dev = tar->dev;
 	if (eh.path)
@@ -740,6 +725,11 @@ restart:
 	tar_offset = tar->offset;
 	ret = erofs_iostream_read(&tar->ios, (void **)&th, sizeof(*th));
 	if (ret != sizeof(*th)) {
+		if (tar->ios.feof) {
+			erofs_warn("unexpected end of file @ %llu (may be non-standard tar without end of archive zeros)", tar_offset);
+			ret = 1;
+			goto out;
+		}
 		if (tar->headeronly_mode || tar->ddtaridx_mode) {
 			ret = 1;
 			goto out;
@@ -832,6 +822,7 @@ out_eot:
 			path[0] = '.';
 			path[1] = '\0';
 		} else {
+			*_path = '\0';
 			while (path[j - 1] == '/')
 				path[--j] = '\0';
 		}
@@ -1004,12 +995,6 @@ out_eot:
 			goto out;
 		}
 
-		if (d->type != EROFS_FT_UNKNOWN) {
-			tarerofs_remove_inode(d->inode);
-			erofs_iput(d->inode);
-		}
-		d->inode = NULL;
-
 		d2 = erofs_rebuild_get_dentry(root, eh.link, tar->aufs,
 					      &dumb, &dumb, false);
 		if (IS_ERR(d2)) {
@@ -1020,14 +1005,23 @@ out_eot:
 			ret = -ENOENT;
 			goto out;
 		}
+		if (d == d2) {
+			ret = 0;
+			goto out;
+		}
 		if (S_ISDIR(d2->inode->i_mode)) {
 			ret = -EISDIR;
 			goto out;
 		}
+
 		inode = erofs_igrab(d2->inode);
+		++inode->i_nlink;
+		if (d->type != EROFS_FT_UNKNOWN) {
+			tarerofs_remove_inode(d->inode);
+			erofs_iput(d->inode);
+		}
 		d->inode = inode;
 		d->type = d2->type;
-		++inode->i_nlink;
 		ret = 0;
 		goto out;
 	} else if (d->type != EROFS_FT_UNKNOWN) {
@@ -1076,7 +1070,7 @@ new_inode:
 		goto out;
 	}
 
-	ret = __erofs_fill_inode(inode, &st, eh.path);
+	ret = __erofs_fill_inode(im, inode, &st, eh.path);
 	if (ret)
 		goto out;
 	inode->i_size = st.st_size;
@@ -1111,8 +1105,8 @@ new_inode:
 								 inode->i_size))
 					ret = -EIO;
 			} else if (tar->try_no_reorder &&
-				   !cfg.c_compr_opts[0].alg &&
-				   !cfg.c_inline_data) {
+				   !sbi->available_compr_algs &&
+				   params->no_datainline) {
 				ret = tarerofs_write_uncompressed_file(inode, tar);
 			} else {
 				ret = tarerofs_write_file_data(inode, tar);
@@ -1122,9 +1116,7 @@ new_inode:
 		}
 		inode->i_nlink++;
 	} else if (!inode->i_nlink) {
-		ret = erofs_init_empty_dir(inode);
-		if (ret)
-			goto out;
+		inode->i_nlink = 2;
 	}
 
 	ret = tarerofs_merge_xattrs(&eh.xattrs, &tar->global.xattrs);
